"""
Main file for ordinary use
"""

from collections import defaultdict
from typing import get_args

import sympy
import pandas as pd

from . import metadata, utils

defaults = metadata.Defaults()
metadata_obj = metadata.Metadata()
_Attributes = metadata.Attributes
_OriginalTables = metadata.OriginalTables
_StandardTables = metadata.StandardTables
_Tables = metadata.Tables


def _check_attribute(attribute: _Attributes | list[_Attributes]) -> None:
    available_attributes = get_args(_Attributes)
    if not isinstance(attribute, (list, tuple)):
        attribute = [attribute]
    for atr in attribute:
        if not atr in available_attributes:
            available_attributes_str = ", ".join(str(x) for x in available_attributes)
            raise KeyError(
                f"Invalid attribute: {atr}. This attribute is not supported."
                f"Available attributes: {available_attributes_str}"
            )


def read_hbs(
    table_name: _Tables,
    from_year: int | None = None,
    to_year: int | None = None,
    add_year: bool | None = None,
    add_duration: bool | None = True,
) -> pd.DataFrame:
    """docs
    """
    if  table_name in metadata.standard_tables:
        table_list: list[_OriginalTables] = metadata_obj.schema[table_name]["table_list"]
    elif table_name in get_args(_OriginalTables):
        table_list = [table_name]
    else:
        raise KeyError

    table = load_table(
        table_name=table_list,
        from_year=from_year,
        to_year=to_year,
        standard=True,
        add_year=add_year,
        add_duration=add_duration,
    )

    table = _imply_table_schema(table, table_name, from_year)
    return table


def load_table(
    table_name: _OriginalTables | list[_OriginalTables] | tuple[_OriginalTables],
    from_year: int | None = None,
    to_year: int | None = None,
    standard: bool | None = None,
    add_year: bool | None = None,
    add_duration: bool | None = None
) -> pd.DataFrame:
    """
    Load Tables
    """
    if standard is None:
        if isinstance(table_name, str):
            standard = table_name in metadata_obj.schema
        else:
            standard = False
    if add_year is None:
        add_year = utils.is_multi_year(table_name, from_year, to_year)

    year_name = utils.create_table_year_product(
        table_name=table_name, from_year=from_year, to_year=to_year
    )
    table_list: list[pd.DataFrame] = []
    for _table_name, year in year_name:
        table = _get_parquet(_table_name, year)
        if add_year:
            table["Year"] = year
        if standard:
            table = _imply_table_schema(table, _table_name, year)
        if add_duration:
            table = _add_duration(table, _table_name)
        table_list.append(table)
    concat_table = pd.concat(table_list, ignore_index=True)

    if not add_year:
        concat_table.attrs["year"] = from_year

    return concat_table


def _get_parquet(
    table_name: str, year: int, download: bool = True, save: bool = True
) -> pd.DataFrame:
    file_name = f"{year}_{table_name}.parquet"
    try:
        table = pd.read_parquet(defaults.processed_data.joinpath(file_name))
    except FileNotFoundError as exc:
        print(
            f"Table {table_name} for year {year} not found at expected location: \n"
            f"{defaults.processed_data.joinpath(file_name)}"
        )
        if download and save:
            _download_parquet(table_name, year)
            table = pd.read_parquet(defaults.processed_data.joinpath(file_name))
        elif download:
            table = pd.read_parquet(
                f"{defaults.online_dir}/parquet_files/{year}_{table_name}.parquet"
            )
        else:
            raise exc
    return table


def _download_parquet(table_name: str, year: int) -> None:
    file_name = f"{year}_{table_name}.parquet"
    file_url = f"{defaults.online_dir}/parquet_files/{file_name}"
    local_path = defaults.processed_data.joinpath(file_name)
    utils.download_file(url=file_url, path=local_path, show_progress_bar=True)


def _imply_table_schema(table, table_name, year: int | None = None):
    """docs
    """
    table = table.copy()
    table_schema = metadata_obj.schema[table_name]

    if "columns" in table_schema:
        instructions = table_schema["columns"]

        for name, instruction in instructions.items():
            if isinstance(year, int):
                instruction = metadata.get_metadata_version(instruction, year)
            table = _apply_column_instruction(table, name, instruction)

    if "order" in table_schema:
        column_order = table_schema["order"]
        table = _order_columns_by_schema(table, column_order)
    return table


def _apply_column_instruction(table, name, instruction):
    if instruction is None:
        pass
    elif instruction["type"] == "categorical":
        table = _apply_categorical_instruction(table, name, instruction)
    elif instruction["type"] == "numerical":
        table = _apply_numerical_instruction(table, name, instruction)

    return table


def _apply_categorical_instruction(table, column_name, instruction):
    categories = instruction["categories"]

    if column_name not in table.columns:
        table[column_name] = None

    # fix me to chack column type and behave accordingly not just changing the type
    table[column_name] = table[column_name].astype(str)

    for category, condition in categories.items():
        if isinstance(condition, str):
            filt = table[column_name] == condition
        elif isinstance(condition, list):
            filt = table[column_name].isin(condition)
        elif isinstance(condition, dict):
            filt = pd.Series(False, index=table.index)
            for other_column, value in condition.items():
                filt = filt | (table[other_column] == value)
        else:
            raise KeyError
        table.loc[filt, column_name] = category

    table[column_name] = table[column_name].astype("category")
    return table


def _apply_numerical_instruction(table, column_name, instruction):
    expr = instruction["expression"]
    pandas_expr = _parse_expression(expr)
    table[column_name] = pd.eval(pandas_expr)
    return table


def _parse_expression(expression, table_name="table"):
    expr = sympy.simplify(expression)
    terms = []

    if len(expr.args) == 0:
        _, var = expr.as_coeff_Mul()
        return f"{table_name}['{var}']"

    if (len(expr.args) == 2) and (len(expr.args[1].args) == 0):
        coeff, var = expr.as_coeff_Mul()
        return f"{table_name}['{var}'] * {coeff}"

    for term in expr.args:
        coeff, var = term.as_coeff_Mul()
        terms.append(f"{table_name}['{var}'].fillna(0) * {coeff}")
    return " + ".join(terms)


def _order_columns_by_schema(table, column_order):
    new_columns = [column for column in column_order if column in table.columns]
    table = table[new_columns]
    return table


def _add_duration(table, table_name):
    table = table.copy()
    default_duration = metadata_obj.commodities["tables"][table_name]["default_duration"]
    table["Duration"] = default_duration
    return table


def add_attribute(
    table: pd.DataFrame,
    attribute: _Attributes | list[_Attributes] | tuple[_Attributes] | None,
    year: int | None = None,
    id_column_name="ID",
    year_column_name: str = "Year",
    attribute_text="names",
) -> pd.DataFrame:
    """docs
    """
    if attribute is None:
        attribute_list = [attr for attr in get_args(_Attributes)]
    elif isinstance(attribute, (list, tuple)):
        attribute_list = [attr for attr in attribute]
    else:
        attribute_list: list[_Attributes] = [attribute]

    table = table.copy()

    for _attribute in attribute_list:
        attribute_column = get_household_attribute(
            _input=table,
            year=year,
            attribute=_attribute,
            id_column_name=id_column_name,
            year_column_name=year_column_name,
            attribute_text=attribute_text,
        )
        table[_attribute] = attribute_column
    return table


def get_household_attribute(
    _input: pd.DataFrame | pd.Series,
    attribute: _Attributes,
    year: int | None = None,
    id_column_name="ID",
    year_column_name: str = "Year",
    attribute_text="names",
) -> pd.Series:
    """docs
    """
    _check_attribute(attribute)
    if isinstance(_input, pd.Series):
        if year is None:
            raise TypeError(
                "Since the input is a Pandas series, the 'year' variable must "
                "be specified. Please provide a year value in the format YYYY."
            )
        return _get_attribute_by_id(_input, year, attribute, attribute_text)
    if not isinstance(_input, pd.DataFrame):
        raise ValueError

    _input = _input.copy()
    years: list[int] = []
    if year is not None:
        years = [year]
        _input["__Year__"] = year
    elif year_column_name in _input.columns:
        years = [int(_year) for _year in _input[year_column_name].unique()]
        _input["__Year__"] = _input[year_column_name]
    elif "year" in _input.attrs:
        year = _input.attrs["year"]
        assert isinstance(year, int)
        years = [year]
        _input["__Year__"] = year
    else:
        raise TypeError(
            "DataFrame does not have a 'year' column. Please provide the "
            "'year' column or specify a value for the 'year' variable."
        )

    attribute_column = pd.Series(None, dtype="object", index=_input.index)
    for _year in years:
        filt = _input["__Year__"] == _year
        id_series = _input.loc[filt, id_column_name]
        attribute_series = _get_attribute_by_id(
            household_id_column=id_series,
            attribute=attribute,
            attribute_text=attribute_text,
            year=_year,
        )
        attribute_column.loc[filt] = attribute_series

    attribute_column = attribute_column.astype("category")
    return attribute_column


def _get_attribute_by_id(
    household_id_column: pd.Series,
    year: int,
    attribute: _Attributes,
    attribute_text="names",
) -> pd.Series:
    attr_dict = metadata_obj.household[attribute]
    text = metadata.get_metadata_version(attr_dict[attribute_text], year)
    attr_codes = _get_attribute_code(household_id_column, year, attribute)
    attr_codes = attr_codes.map(text)
    attr_codes = attr_codes.astype("category")
    return attr_codes


def _get_attribute_code(
    household_id_column: pd.Series,
    year: int,
    attribute: _Attributes,
) -> pd.Series:
    id_length = metadata.get_metadata_version(metadata_obj.household["ID_Length"], year)
    attr_dict = metadata_obj.household[attribute]
    position = metadata.get_metadata_version(attr_dict["position"], year)
    start, end = position["start"], position["end"]
    attr_codes = household_id_column % pow(10, (id_length - start))
    attr_codes = attr_codes // pow(10, (id_length - end))
    return attr_codes


def add_classification(
    table: pd.DataFrame,
    classification: str = "original",
    level: int | list[int] | None = None,
    year: int | None = None,
    code_column_name: str = "Code",
    year_column_name: str = "Year",
    new_column_name: str | list[str] | None = None,
) -> pd.DataFrame:
    """docs
    """
    table = table.copy()

    if level is None:
        levels = metadata_obj.commodities[classification]["default_levels"]
    elif isinstance(level, int):
        levels = [level]
    elif isinstance(level, list):
        levels = level
    else:
        raise TypeError

    if new_column_name is None:
        if "default_names" in metadata_obj.commodities[classification]:
            column_names = metadata_obj.commodities[classification]["default_names"]
        else:
            column_names = [f"{classification}-{_level}" for _level in levels]
    elif isinstance(new_column_name, str):
        column_names = [new_column_name]
    else:
        column_names = new_column_name

    assert len(levels) == len(column_names)

    level_and_name = zip(levels, column_names)

    for _level, column_name in level_and_name:
        classification_column = get_code_classification(
            table,
            classification=classification,
            level=_level,
            year=year,
            code_column_name=code_column_name,
            year_column_name=year_column_name,
        )
        table[column_name] = classification_column

    return table


def get_code_classification(
    _input: pd.DataFrame | pd.Series,
    classification: str,
    level: int,
    year: int | None = None,
    code_column_name: str = "Code",
    year_column_name: str = "Year",
) -> pd.Series:
    """docs
    """
    if isinstance(_input, pd.Series):
        if year is None:
            raise TypeError(
                "Since the input is a Pandas series, the 'year' variable must "
                "be specified. Please provide a year value in the format YYYY."
            )
        return _get_classification_by_code(_input, classification, level, year)
    if not isinstance(_input, pd.DataFrame):
        raise ValueError

    _input = _input.copy()
    if year is not None:
        years = [year]
        _input["__Year__"] = year
    elif year_column_name in _input.columns:
        years = [int(y) for y in _input[year_column_name].unique()]
        _input["__Year__"] = _input[year_column_name]
    elif "year" in _input.attrs:
        year = _input.attrs["year"]
        if year is not None:
            years = [year]
        else:
            raise KeyError
        _input["__Year__"] = year
    else:
        raise TypeError(
            "DataFrame does not have a 'year' column. Please provide the "
            "'year' column or specify a value for the 'year' variable."
        )

    classification_column = pd.Series(None, dtype="object", index=_input.index)
    for _year in years:
        filt = _input["__Year__"] == _year
        code_series = _input.loc[filt, code_column_name]
        classification_series = _get_classification_by_code(
            commodity_code_column=code_series,
            classification=classification,
            level=level,
            year=_year,
        )
        classification_column.loc[filt] = classification_series

    classification_column = classification_column.astype("category")
    return classification_column


def _get_classification_by_code(
    commodity_code_column: pd.Series,
    classification: str,
    level: int,
    year: int,
) -> pd.Series:
    translator = _build_translator(
        classification=classification, level=level, year=year
    )
    classification_column = commodity_code_column.map(translator)
    classification_column = classification_column.astype("category")
    return classification_column


def _build_translator(
    classification: str,
    level: int,
    year: int,
    attribute: str = "name",
    default_value: str | None = None,
) -> dict:
    def closure(_input):
        def inner_function():
            return _input
        return inner_function

    commodity_codes = metadata_obj.commodities[classification]["items"]
    commodity_codes = metadata.get_metadata_version(commodity_codes, year)
    selected_items = {
        name: info for name, info in commodity_codes.items() if info["level"] == level
    }
    translator = {}
    if attribute == "name":
        for name, info in selected_items.items():
            categories = metadata.get_categories(info)
            for category_info in categories:
                if "default" in category_info:
                    translator = defaultdict(closure(name), translator)
                    break
                code_range = _get_code_range(category_info["code"])
                for code in code_range:
                    translator[code] = name
    else:
        for info in selected_items.values():
            categories = metadata.get_categories(info)
            for category_info in categories:
                try:
                    attribute_value = category_info[attribute]
                except KeyError:
                    attribute_value = default_value
                if "default" in category_info:
                    translator = defaultdict(closure(attribute_value), translator)
                    break
                code_range = _get_code_range(category_info["code"])
                for code in code_range:
                    translator[code] = attribute_value

    return translator


def _get_code_range(code_range_info: int | dict | list) -> list[int]:
    if isinstance(code_range_info, int):
        code_range = [code_range_info]
    elif isinstance(code_range_info, dict):
        if ("start" in code_range_info) and ("end" in code_range_info):
            code_range = list(range(code_range_info["start"], code_range_info["end"]))
        elif "code" in code_range_info:
            code_range = _get_code_range(code_range_info["code"])
        else:
            raise KeyError
    elif isinstance(code_range_info, list):
        code_range = []
        for element in code_range_info:
            code_range.extend(_get_code_range(element))
    else:
        raise KeyError

    return code_range



"""
The purpose of the archive handler module is to allow downloading and storing
raw data more easily.
"""

import shutil
import subprocess
import platform
from pathlib import Path

from tqdm import tqdm
import pandas as pd
import pyodbc

from . import metadata, utils


defaults = metadata.Defaults()


def setup(
        from_year: int | None = None, to_year: int | None = None,
        replace: bool = False) -> None:
    """
    Download census data archive files, unpack them, and extract tables from the MS Access files
    for each year specified in the range of years and save them as CSV.

    Parameters
    ----------
    from_year : int, optional
        The starting year of the range. If not provided, files will be set up for all available
        years.
    to_year : int, optional
        The ending year of the range. If provided, files will be set up for the years between
        from_year and to_year, inclusive. If not provided, files will be set up for the single
        year specified in from_year.
    replace : bool, optional
        If True, existing files will be overwritten. If False, existing files will be skipped.
        The default is False.

    Returns
    -------
    None

    See Also
    --------
    download : A function that downloads household census data archive files.
    unpack : A function that unpacks household census data archive files.
    extract_tables : A function that extracts tables from a census MS Access file and saves them
        as CSV files

    Notes
    -----

    .. note::

        For ordinary usage, it is not necessary to download these files. This data is intended
        for use during development and metadata writing.

    .. warning::
        Note that setting up files for the entire range of years will occupy approximately 12 GB
        of storage.
    

    Examples
    --------
    To set up all available files:    

    >>> setup()

    To set up files for the year 1393:    

    >>> setup(1393)

    To set up files for the years 1370 to 1380, inclusive:    

    >>> setup(1370, 1380)
    """
    download(from_year, to_year, replace)
    unpack(from_year, to_year, replace)
    extract_tables(from_year, to_year, replace)


def download(
        from_year: int | None = None, to_year: int | None = None,
        replace: bool = False) -> None:
    """
    Downloads household census data archive files.
    
    Parameters
    ----------
    from_year : int, optional
        The starting year of the range. If provided, download files for that specific year only.
        If not provided, download files for all available years.
    to_year : int, optional
        The ending year of the range. If provided, download files for the years between
        from_year and to_year, inclusive. If not provided, and from_year is provided,
        download files for the single year specified in from_year.
    replace : bool, optional
        If True, overwrite existing files. If False, skip existing files. Default is False.
        
    Returns
    -------
    None

    See Also
    --------
    setup : A function that downloads, unpacks and extracts all tables from household census data
        archive files and save them as CSV.

    Examples
    --------
    To download all available files:    

    >>> download()

    To download files for the year 1393:    

    >>> download(1393)

    To download files for the years 1370 to 1380, inclusively:    

    >>> download(1370, 1380)
    """
    from_year, to_year = utils.build_year_interval(from_year, to_year)
    Path(defaults.archive_files).mkdir(exist_ok=True, parents=True)
    for year in range(from_year, to_year):
        _download_year_file(year, replace=replace)


def _download_year_file(year: int, replace: bool = True) -> None:
    """
    Downloads the household census data archive file for a specified year
    from the online directory in its original form.

    :param year: The year for which to download the household census file.
    :type year: int
    :param replace: Whether to replace the file if it already exists.
    Defaults to True.
    :type replace: bool, optional
    :return: None, as this function does not return anything.

    """
    file_name = f"{year}.rar"
    file_url = f"{defaults.online_dir}/original_files/{file_name}"
    defaults.archive_files.mkdir(parents=True, exist_ok=True)
    local_path = defaults.archive_files.joinpath(file_name)
    if (not Path(local_path).exists()) or replace:
        utils.download_file(url=file_url, path=local_path,
                            show_progress_bar=True)


def _unpack_archive_with_7zip(
        compressed_file_path: str, output_directory: str) -> None:
    """
    Extracts the contents of a compressed file using the 7-Zip tool.

    :param compressed_file_path: The path to the compressed file to be
    extracted.
    :type compressed_file_path: str
    :param output_directory: The path to the directory where the
    extracted files will be stored.
    :type output_directory: str
    :return: None, as this function does not return anything.

    """
    if platform.system() == "Windows":
        seven_zip_file_path = Path().joinpath("7-Zip", "7z.exe")
        if not seven_zip_file_path.exists():
            utils.download_7zip()
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file_path,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
            shell=True,
        )
    elif platform.system() == "Linux":
        seven_zip_file_path = defaults.pack_dir.joinpath("7-Zip", "7zz")
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file_path,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
        )


def unpack(
        from_year: int | None = None, to_year: int | None = None,
        replace: bool = False) -> None:
    """
    Unpacks census data archive files, extracts tables from the MS Access files and saves them
    as CSV files for each year specified in the range of years.

    Parameters
    ----------
    from_year : int, optional
        The starting year of the range. If not provided, unpack files for all available years.
    to_year : int, optional
        The ending year of the range. If provided, unpack files for the years between from_year
        and to_year, inclusive. unpack files for the single year specified in from_year.
    replace : bool, optional
        If True, overwrite existing files. If False, skip existing files. Default is False.

    Returns
    -------
    None

    See Also
    --------
    setup : A function that downloads, unpacks and extracts all tables from household census data
        archive files and save them as CSV.

    Notes
    -----
    .. note::

        Any nested archive files within the extracted directories will also be extracted
        recursively.

    Examples
    --------
    To unpack all available archives:    

    >>> unpack()

    To unpack archives from 1393 to 1400:    

    >>> unpack(1393, 1400)
    """
    from_year, to_year = utils.build_year_interval(
        from_year=from_year, to_year=to_year)
    for year in tqdm(range(from_year, to_year), desc="Unziping raw data", unit="file"):
        _unpack_yearly_data_archive(year, replace=replace)


def _unpack_yearly_data_archive(year: int, replace: bool = True):
    """
    Extracts data archive for the given year.
    """
    file_path = defaults.archive_files.joinpath(f"{year}.rar")
    year_directory = defaults.unpacked_data.joinpath(str(year))
    if year_directory.exists():
        if not replace:
            return
        shutil.rmtree(year_directory)
    year_directory.mkdir(parents=True)
    _unpack_archive_with_7zip(file_path, year_directory)
    _unpack_archives_recursive(year_directory)
    _remove_created_directories(year_directory)


def _unpack_archives_recursive(directory: str | Path):
    """
    Extract all archives with the extensions ".zip" and ".rar" found
    recursively in the given directory and its subdirectories using
    7zip.

    :param directory: The directory path to search for archives.
    :type directory: str
    :return: None

    """
    while True:
        all_files = list(Path(directory).iterdir())
        archive_files = [
            file for file in all_files if file.suffix in (".zip", ".rar")]
        if len(archive_files) == 0:
            break
        for file in archive_files:
            _unpack_archive_with_7zip(file, directory)
            Path(file).unlink()


def _remove_created_directories(directory: Path):
    for path in directory.iterdir():
        if path.is_dir():
            shutil.rmtree(path)


def extract_tables(
        from_year: int | None = None, to_year: int | None = None,
        replace: bool = False) -> None:
    """
    Extracts tables from a census MS Access file and saves them as CSV files.

    Parameters
    ----------
    from_year : int, optional
        The starting year of the range. If not provided, extract tables for all available years.
    to_year : int, optional
        The ending year of the range. If provided, extract tables for the years between from_year
        and to_year, inclusive. extract tables for the single year specified in from_year.
    replace : bool, optional
        If True, overwrite existing files. If False, skip existing files. Default is False.

    Returns
    -------
    None

    See Also
    --------
    setup : A function that downloads, unpacks and extracts all tables from household census data
        archive files and save them as CSV.

    Examples
    --------
    To extract tables from all available census files:    

    >>> extract_tables()

    To extract tables from all available census files from 1393 to 1400:    

    >>> extract_tables(from_year=1393, to_year=1400)
    """
    from_year, to_year = utils.build_year_interval(from_year, to_year)
    for year in range(from_year, to_year):
        _extract_tables_from_access_file(year, replace)


def _extract_tables_from_access_file(year: int, replace: bool = True) -> None:
    table_list = _get_access_table_list(year)
    for table_name in tqdm(
        table_list, desc=f"Extracting data from {year}", unit="table"
    ):
        _extract_table(year, table_name, replace)


def _get_access_table_list(year: int) -> list:
    connection_string = _make_connection_string(year)
    with pyodbc.connect(connection_string) as connection:
        cursor = connection.cursor()
        access_tables = cursor.tables()
        table_list = []
        for table in access_tables:
            table_list.append(table.table_name)
    table_list = [table for table in table_list if table.find("MSys") == -1]
    return table_list


def _extract_table(year: int, table_name: str, replace: bool = True):
    file_name = _change_1380_table_names(year, table_name)
    year_directory = defaults.extracted_data.joinpath(str(year))
    year_directory.mkdir(parents=True, exist_ok=True)
    file_path = year_directory.joinpath(f"{file_name}.csv")
    if (file_path.exists()) and (not replace):
        return

    try:
        table = _get_access_table(year, table_name)
    except (pyodbc.ProgrammingError, pyodbc.OperationalError):
        tqdm.write(f"table {table_name} from {year} failed to extract")
        return
    table.to_csv(file_path, index=False)


def _change_1380_table_names(year: int, table_name: str):
    if year == 1380:
        if table_name in metadata.Metadata.other["unusual_names_of_1380"]:
            table_name = metadata.Metadata.other["unusual_names_of_1380"][table_name]
    return table_name


def _get_access_table(year: int, table_name: str) -> pd.DataFrame:
    connection_string = _make_connection_string(year)
    with pyodbc.connect(connection_string) as connection:
        cursor = connection.cursor()
        rows = cursor.execute(f"SELECT * FROM [{table_name}]").fetchall()
        headers = [c[0] for c in cursor.description]
    table = pd.DataFrame.from_records(rows, columns=headers)
    return table


def _make_connection_string(year: int):
    if platform.system() == "Windows":
        driver = "Microsoft Access Driver (*.mdb, *.accdb)"
    else:
        driver = "MDBTools"

    year_directory = defaults.unpacked_data.joinpath(str(year))
    access_file_path = _find_access_file_by_extension(year_directory)
    conn_str = f"DRIVER={{{driver}}};" f"DBQ={access_file_path};"
    return conn_str


def _find_access_file_by_extension(directory: Path) -> Path:
    files = list(directory.iterdir())
    for file in files:
        if file.suffix.lower() in [".mdb", ".accdb"]:
            return file
    raise FileNotFoundError



"""
Utility functions
"""

import subprocess
from pathlib import Path
import platform
from zipfile import ZipFile

from tqdm import tqdm
import requests

from . import metadata


defaults = metadata.Defaults()
metadata_obj = metadata.Metadata()
_Tables = metadata.Tables


def download_file(
    url: str, path: str | Path | None = None, show_progress_bar: bool = False
) -> Path:
    """
    Download a file from a URL to a local path.

    :param url: The URL to download the file from.
    :param path: The local path to save the downloaded file to. If None, a
        temporary directory is used.
    :param show_progress_bar: Whether to display a progress bar during the
        download.
    :param timeout: The timeout for the download request, in seconds.
    :returns: The local path of the downloaded file.

    """
    if isinstance(path, str):
        path = Path(path)
        file_name = path.name
    elif path is None:
        temp_folder = defaults.pack_dir.joinpath("temp")
        temp_folder.mkdir(exist_ok=True)
        file_name = url.split("/")[-1]
        path = temp_folder.joinpath(file_name)
    else:
        file_name = path.name

    response = requests.get(url, timeout=10, stream=True)
    file_size = response.headers.get("content-length")
    if file_size is not None:
        file_size = int(file_size)
    else:
        raise FileNotFoundError
    download_bar = tqdm(
        desc=f"downloading {file_name}",
        total=file_size,
        unit="B",
        unit_scale=True,
        disable=not show_progress_bar,
    )
    Path(path.parent).mkdir(parents=True, exist_ok=True)
    with open(path, mode="wb") as file:
        for chunk in response.iter_content(chunk_size=4096):
            download_bar.update(len(chunk))
            file.write(chunk)
    download_bar.close()
    return path


def download_7zip():
    """
    Download the appropriate version of 7-Zip for the current operating system
    and architecture, and extract it to the root directory.

    """
    print(
        f"Downloading 7-Zip for {platform.system()} with {platform.architecture()[0]} architecture"
    )
    file_name = f"{platform.system()}-{platform.architecture()[0]}.zip"
    file_path = Path().joinpath("temp", file_name)
    file_path = download_file(
        f"{defaults.online_dir}/7-Zip/{file_name}",
        show_progress_bar=True,
    )
    with ZipFile(file_path) as zip_file:
        zip_file.extractall(defaults.pack_dir)
    file_path.unlink()

    if platform.system() == "Linux":
        defaults.pack_dir.joinpath("7-Zip", "7zz").chmod(0o771)


def extract_with_7zip(compressed_file_path: str, output_directory: str) -> None:
    """
    Extracts the contents of a compressed file using the 7-Zip tool.

    :param compressed_file_path: The path to the compressed file to be extracted.
    :type compressed_file_path: str
    :param output_directory: The path to the directory where the extracted files will be stored.
    :type output_directory: str
    :return: None, as this function does not return anything.

    """
    if platform.system() == "Windows":
        seven_zip_file_path = Path().joinpath("7-Zip", "7z.exe")
        if not seven_zip_file_path.exists():
            download_7zip()
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file_path,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
            shell=True,
        )
    elif platform.system() == "Linux":
        seven_zip_file_path = defaults.pack_dir.joinpath("7-Zip", "7zz")
        if not seven_zip_file_path.exists():
            download_7zip()
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file_path,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
        )


def interpret_number_range(
    number_range_info: int | dict | list,
    default_end: int | None = None,
    keyword: str = "code",
) -> list[int]:
    """
    A tool for interpretation of number ranges
    """
    number_range: list[int] = []
    if isinstance(number_range_info, int):
        number_range.append(number_range_info)
    elif isinstance(number_range_info, dict):
        if ("start" in number_range_info) and ("end" in number_range_info):
            number_range.extend(
                list(range(number_range_info["start"], number_range_info["end"]))
            )
        elif ("start" in number_range_info) and (default_end is not None):
            number_range.extend(list(range(number_range_info["start"], default_end)))
        elif keyword in number_range_info:
            number_range.extend(
                interpret_number_range(number_range_info[keyword], default_end=default_end)
            )
        else:
            raise KeyError
    elif isinstance(number_range_info, list):
        for element in number_range_info:
            number_range.extend(interpret_number_range(element, default_end=default_end))
    else:
        raise KeyError

    return number_range


def build_year_interval(
    from_year: int | None,
    to_year: int | None,
    earliest_year: int | None = None,
    latest_year: int | None = None,
) -> tuple[int, int]:
    """
    Returns a tuple of two integers representing a range of years.

    :param from_year: An integer representing the starting year of the range.
        If `None`, the default starting year is used.
    :type from_year: int or None

    :param to_year: An integer representing the ending year of the range. If
        `None`, the default ending year + 1 is used.
    :type to_year: int or None

    :return: A tuple representing the range of years.
    :rtype: tuple of two integers

    :raises: ValueError if `from_year` is greater than `to_year`.

    Example:

    >>> build_year_interval()
    (1363, 1401)

    >>> build_year_interval(1390)
    (1390, 1401)

    >>> build_year_interval(1380, 1385)
    (1380, 1386)

    >>> build_year_interval(1383, 1375)
    ValueError: `from_year` must be less than `to_year`.
    """
    if earliest_year is None:
        earliest_year = defaults.first_year
    if latest_year is None:
        latest_year = defaults.last_year

    if (from_year is not None) and (to_year is not None):
        if to_year < from_year:
            raise ValueError("`from_year` is greater than `to_year`")
        _from_year, _to_year = from_year, to_year + 1
    elif (from_year is not None) and (to_year is None):
        _from_year, _to_year = from_year, from_year + 1
    elif (from_year is None) and (to_year is not None):
        raise KeyError
    else:
        _from_year, _to_year = earliest_year, latest_year + 1

    return _from_year, _to_year


def build_year_interval_for_table(
    table_name: _Tables, from_year: int | None = None, to_year: int | None = None
) -> list[int]:
    """_summary_

    Parameters
    ----------
    table_name : str
        _description_
    from_year : int | None, optional
        _description_, by default None
    to_year : int | None, optional
        _description_, by default None

    Returns
    -------
    _type_
        _description_
    """
    availability_info = metadata_obj.tables["yearly_table_availability"][table_name]
    available_years = interpret_number_range(availability_info, default_end=defaults.last_year+1)
    from_year, to_year = build_year_interval(from_year, to_year)
    selected_years = [year for year in range(from_year, to_year) if year in available_years]
    return selected_years


def create_table_year_product(
    table_name: _Tables | list[_Tables] | tuple[_Tables],
    from_year: int | None = None,
    to_year: int | None = None,
) -> list[tuple[_Tables, int]]:
    """_summary_

    Parameters
    ----------
    table_name : str | List[str]
        _description_
    from_year : int | None
        _description_
    to_year : int | None, optional
        _description_, by default None

    Returns
    -------
    List[tuple]
        _description_
    """
    if isinstance(table_name, str):
        table_list: list[_Tables] = [table_name]
    else:
        table_list: list[_Tables] = list(table_name)

    product_list = []
    for _table_name in table_list:
        years = build_year_interval_for_table(_table_name, from_year, to_year)
        product_list.extend([(_table_name, year) for year in years])
    return product_list


def is_multi_year(
    table_name: _Tables | list[_Tables] | tuple[_Tables],
    from_year: int | None = None,
    to_year: int | None = None,
) -> bool:
    if isinstance(table_name, str):
        table_list: list[_Tables] = [table_name]
    else:
        table_list = [table for table in table_name]

    for _table_name in table_list:
        years = build_year_interval_for_table(_table_name, from_year, to_year)
        if len(years) > 1:
            return True
    return False



"""
Module for cleaning raw data into proper format
"""

from pathlib import Path
from typing import Hashable

from tqdm import tqdm
import pandas as pd
import numpy as np

from . import metadata, utils


defaults = metadata.Defaults()
metadata_obj = metadata.Metadata()
_Tables = metadata.Tables


def load_table_data(
    table_name: str, year: int, urban: bool | None = None
) -> pd.DataFrame:
    """
    Reads CSV file(s) containing data for a given table, year, and urban/rural category,
    and returns the data as a pandas DataFrame.

    :param table_name: The name of the table to be read.
    :type table_name: str
    :param year: The year of the data to be read.
    :type year: int
    :param urban: A boolean indicating whether to read data for urban areas only (`True`),
        rural areas only (`False`), or both (`None`, default). If `None`, data for both
        urban and rural areas will be read.
    :type urban: bool|None
    :return: A DataFrame containing the concatenated data from the CSV file(s).
    :rtype: pd.DataFrame
    :raises FileNotFoundError: If the CSV file(s) cannot be found at the expected file path(s).
    :raises ValueError: If the table name or year are invalid, or if the table metadata file is
        corrupt.
    :raises pd.errors.EmptyDataError: If the CSV file(s) are empty.

    :example:

    >>> load_table_data("food", 1393, urban=True)
    """

    if urban is None:
        urban_stats = [True, False]
    else:
        urban_stats = [urban]

    tables = []
    for is_urban in urban_stats:
        file_path = _build_file_path(table_name, year, is_urban)
        tables.append(pd.read_csv(file_path, low_memory=False))
    table = pd.concat(tables, axis="index", ignore_index=True)
    return table


def _build_file_path(table_name: str, year: int, is_urban: bool) -> Path:
    urban_rural = "U" if is_urban else "R"
    year_string = year % 100 if year < 1400 else year
    table_metadata = _get_table_metadata(table_name, year, is_urban)
    file_code = metadata.get_metadata_version(table_metadata["file_code"], year)
    file_name = f"{urban_rural}{year_string}{file_code}.csv"
    file_path = defaults.extracted_data.joinpath(str(year), file_name)
    return file_path


def _get_table_metadata(
    table_name: str, year: int, is_urban: bool | None = None
) -> dict:
    table_metadata = metadata_obj.tables[table_name]
    table_metadata = metadata.get_metadata_version(table_metadata, year)

    if is_urban is True:
        if "urban" in table_metadata:
            table_metadata = table_metadata["urban"]
    if is_urban is False:
        if "rural" in table_metadata:
            table_metadata = table_metadata["rural"]

    table_metadata["table_name"] = table_name
    table_metadata["year"] = year
    table_metadata["is_urban"] = is_urban

    return table_metadata


def clean_table_with_metadata(table_name: str, year: int) -> pd.DataFrame:
    """
    Clean the specified table using metadata and return a cleaned pandas DataFrame.

    :param table_name: The name of the table to be cleaned.
    :type table_name: str
    :param year: The year for which the table will be cleaned.
    :type year: int
    :return: A cleaned DataFrame with the specified table's data.
    :rtype: pandas.DataFrame

    """
    cleaned_table_list = []
    for is_urban in [True, False]:
        table = load_table_data(table_name, year, is_urban)
        table_metadata = _get_table_metadata(table_name, year, is_urban)
        cleaned_table = _apply_metadata_to_table(table, table_metadata)
        cleaned_table_list.append(cleaned_table)
    final_table = pd.concat(cleaned_table_list, ignore_index=True)
    return final_table


def _apply_metadata_to_table(table: pd.DataFrame, table_metadata: dict) -> pd.DataFrame:
    cleaned_table = pd.DataFrame()
    for column_name, column in table.items():
        assert isinstance(column_name, str)
        column_metadata = _get_column_metadata(table_metadata, column_name.upper())
        if column_metadata == "drop":
            continue
        column = _apply_metadata_to_column(column, column_metadata)
        cleaned_table[column_metadata["new_name"]] = column
    return cleaned_table


def _get_column_metadata(table_metadata: dict, column_name: Hashable) -> dict:
    table_settings = _get_table_settings(table_metadata)
    year = table_metadata["year"]
    columns_metadata = table_metadata["columns"]
    columns_metadata = metadata.get_metadata_version(columns_metadata, year)
    try:
        column_metadata = columns_metadata[column_name]
    except KeyError:
        column_metadata = table_settings["missings"]
    else:
        column_metadata = metadata.get_metadata_version(column_metadata, year)
    return column_metadata


def _get_table_settings(table_metadata: dict) -> dict:
    default_table_settings = metadata_obj.tables["default_table_settings"]
    try:
        table_settings = table_metadata["settings"]
    except KeyError:
        return default_table_settings
    for key in default_table_settings:
        if key not in table_settings:
            table_settings[key] = default_table_settings[key]
    return table_settings


def _apply_metadata_to_column(column: pd.Series, column_metadata: dict) -> pd.Series:
    column = _apply_type_to_column(column, column_metadata)
    return column


def _apply_type_to_column(column: pd.Series, column_metadata: dict) -> pd.Series:
    column = _convert_empty_items_to_nan(column)
    non_empty = column.notna()
    new_column = pd.Series(np.nan, index=column.index)
    if column_metadata["type"] == "boolian":
        new_column = column.astype("Int32").astype("category")
        new_column.loc[non_empty] = new_column == column_metadata["true_condition"]
    elif column_metadata["type"] in ("unsigned", "integer", "float"):
        new_column = pd.to_numeric(column, downcast=column_metadata["type"])
    elif column_metadata["type"] == "category":
        new_column = column.astype("Int32").astype("category")
        new_column = new_column.cat.rename_categories(column_metadata["categories"])
    elif column_metadata["type"] == "string":
        new_column = column.copy()
    return new_column


def _convert_empty_items_to_nan(column: pd.Series):
    if pd.api.types.is_numeric_dtype(column):
        return column
    chars_to_remove = r"\n\r\,\@\-\+\*"
    column = column.str.replace(f"[{chars_to_remove}]*", "", regex=True)
    column = column.replace(r"\A\s*\Z", np.nan, regex=True)
    return column


def parquet_clean_data(
    table_name: _Tables | list[_Tables],
    from_year: int | None = None,
    to_year: int | None = None,
) -> None:
    """
    Clean and process data for a specified table and year range, and save it in
    Parquet format.

    :param table_name: Name of the table to be cleaned and processed.
    :type table_name: str

    :param from_year: Starting year of the data to be cleaned and processed
        (inclusive). If not specified, defaults to the earliest year available
        for the table.
    :type from_year: int or None

    :param to_year: Ending year of the data to be cleaned and processed
        (inclusive). If not specified, defaults to the last year.
    :type to_year: int or None

    :return: None

    :raises FileNotFoundError: If the function is unable to find the specified
        table(s) CSV file(s).

    :raises Exception: If an error occurs during the cleaning and processing of
        the data.

    .. note:: The cleaned and processed data will be saved in Parquet format to
        the `processed_data` directory in the project's default settings.

    .. seealso:: `utils.build_year_interval`, `clean_table_with_metadata`

    """
    table_year = utils.create_table_year_product(table_name, from_year, to_year)
    pbar = tqdm(total=len(table_year), desc="Preparing ...", unit="Table")
    for _table_name, year in table_year:
        pbar.update()
        pbar.desc = f"Table: {_table_name}, Year: {year}"
        table = clean_table_with_metadata(_table_name, year)
        Path(defaults.processed_data).mkdir(exist_ok=True)
        table.to_parquet(
            defaults.processed_data.joinpath(f"{year}_{_table_name}.parquet"),
            index=False,
        )
    pbar.close()



"""
DataFrame extention
"""
import pandas as pd

from .data_engine import add_attribute, add_classification
from .metadata import (
    Attributes as _Attributes,
)

@pd.api.extensions.register_dataframe_accessor("hbs")
class HBSDF:
    """Household Budget Survay pandas API"""

    def __init__(self, data_table):
        self.table: pd.DataFrame = data_table
        self.year: int | None = None
        self.id_column_name: str = "ID"
        self.year_column_name: str = "Year"
        self.code_column_name: str = "Code"

    def add_attribute(self, attribute: _Attributes | list[_Attributes]) -> pd.DataFrame:
        """
        Add Household attributes to the DataFrame
        """
        return add_attribute(
            self.table,
            attribute=attribute,
            attribute_text="names",
            year=self.year,
            id_column_name=self.id_column_name,
            year_column_name=self.year_column_name,
        )

    def add_classification(
        self,
        classification: str = "original",
        level: int | list[int] | None = None,
        new_column_name: str | list[str] | None = None,
        ) -> pd.DataFrame:
        """
        Add Commodities Classification to the DataFrame
        """
        return add_classification(
            self.table,
            classification=classification,
            level=level,
            new_column_name=new_column_name,
            code_column_name=self.code_column_name,
            year_column_name=self.year_column_name,
        )



"""
Metadata module
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Literal

import yaml


PACKAGE_DIRECTORY = Path(__file__).parent
ROOT_DIRECTORT = Path().absolute()


Attributes = Literal["Urban-Rural", "Province", "Region"]

GeneralTables = Literal[
    "household_information",
    "members_properties",
    "house_specifications",
]

ExpenditureTables = Literal[
    "food",
    "tobacco",
    "cloth",
    "home",
    "furniture",
    "medical",
    "transportation",
    "communication",
    "entertainment",
    "education",
    "hotel",
    "other",
    "durable",
    "investment",
]

IncomeTables = Literal[
    "employment_income",
    "self_employed_income",
    "other_income",
    "subsidy",
    "public_employment_income",
    "private_employment_income",
]

OriginalTables = Literal[GeneralTables, ExpenditureTables, IncomeTables]

standard_tables = (
    "Expenditures",
)
StandardTables = Literal[
    "Expenditures",
]

Tables = Literal[OriginalTables, StandardTables]


def open_yaml(path):
    """
    Read the contents of a YAML file relative to the root directory and return it as a dictionary.

    :param path: The path to the YAML file, relative to the root directory.
    :type path: str

    :return: The contents of the YAML file as a dictionary.
    :rtype: dict

    :raises yaml.YAMLError: If there is an error parsing the YAML file.

    """
    path = PACKAGE_DIRECTORY.joinpath(path)
    with open(path, mode="r", encoding="utf8") as yaml_file:
        yaml_content = yaml.load(yaml_file, Loader=yaml.CLoader)
    return yaml_content


@dataclass
class Metadata:
    """
    A dataclass for accessing metadata used in other parts of the project.

    """

    tables = open_yaml("metadata/tables.yaml")
    maps = open_yaml("metadata/maps.yaml")
    household = open_yaml("metadata/household.yaml")
    commodities = open_yaml("metadata/commodities.yaml")
    schema = open_yaml("metadata/schema.yaml")
    other = open_yaml("metadata/other.yaml")


@dataclass
class Defaults:
    """
    This dataclass provides access to default values that are used in other
    parts of the project. It first attempts to load the settings from a local
    `hbsir-settings.yaml` file in the root directory. If the file is not found,
    it loads the settings from the sample file named `settings-sample.yaml`
    located in config folder in library directory.

    """

    if ROOT_DIRECTORT.joinpath("hbsir-settings.yaml").exists():
        settings = open_yaml(ROOT_DIRECTORT.joinpath("hbsir-settings.yaml"))
    elif PACKAGE_DIRECTORY.joinpath("config", "hbsir-settings.yaml").exists():
        settings = open_yaml(PACKAGE_DIRECTORY.joinpath("config", "hbsir-settings.yaml"))
    elif PACKAGE_DIRECTORY.joinpath("config", "settings-sample.yaml").exists():
        settings = open_yaml(PACKAGE_DIRECTORY.joinpath("config", "settings-sample.yaml"))
    else:
        raise FileNotFoundError

    # online directory
    online_dir = settings["online_directory"]

    # local directory
    pack_dir: Path = PACKAGE_DIRECTORY
    root_dir: Path = ROOT_DIRECTORT
    if Path(settings["local_directory"]).is_absolute():
        local_dir = Path(settings["local_directory"])
    elif settings["in_root"]:
        local_dir: Path = root_dir.joinpath(settings["local_directory"])
    else:
        local_dir: Path = pack_dir.joinpath(settings["local_directory"])
    archive_files: Path = local_dir.joinpath(settings["archive_files"])
    unpacked_data: Path = local_dir.joinpath(settings["unpacked_data"])
    extracted_data: Path = local_dir.joinpath(settings["extracted_data"])
    processed_data: Path = local_dir.joinpath(settings["processed_data"])

    first_year: int = settings["first_year"]
    last_year: int = settings["last_year"]


def get_latest_version_year(metadata_dict: dict, year: int) -> int | bool:
    """
    Retrieve the most recent available version of metadata that matches or
    precedes the given year, provided that the metadata is versioned.

    :param metadata_dict: A dictionary representing the metadata.
    :type metadata_dict: dict

    :param year: The year to which the version of the metadata should match or
        precede.
    :type year: int

    :return: The version number of the most recent metadata version that
        matches or precedes the given year, or None if the metadata is not
        versioned.
    :rtype: int or None

    """
    if not isinstance(metadata_dict, dict):
        return False
    if "versions" in metadata_dict:
        return True
    version_list = list(metadata_dict.keys())
    for element in version_list:
        if not isinstance(element, int):
            return False
        if (element < 1300) or (element > 1500):
            return False

    selected_version = 0
    for version in version_list:
        if version <= year:
            selected_version = max(selected_version, version)
    return selected_version


def get_metadata_version(metadata_dict: dict, year: int) -> dict:
    """
    Retrieve the metadata version that matches or precedes the given year,
    returning the complete metadata for that version.

    :param metadata_dict: A dictionary representing the metadata.
    :type metadata_dict: dict

    :param year: The year to which the version of the metadata should match or
        precede.
    :type year: int

    :return: A dictionary containing the complete metadata for the version that
        matches or precedes the given year. If the metadata is not versioned, the
        function returns the original metadata dictionary.
    :rtype: dict

    """
    selected_version = get_latest_version_year(metadata_dict, year)
    if selected_version is True:
        selected_version = get_latest_version_year(metadata_dict["versions"], year)
        metadata_version = {key: value for key, value in metadata_dict.items() if key != "versions"}
        for key, value in metadata_dict["versions"][selected_version].items():
            metadata_version[key] = value

    elif selected_version is False:
        metadata_version =  metadata_dict

    else:
        metadata_version = metadata_dict[selected_version]

    return metadata_version


def get_categories(metadata_dict: dict) -> list:
    """_summary_

    Parameters
    ----------
    metadata_dict : dict
        _description_

    Returns
    -------
    _type_
        _description_
    """
    if "categories" not in metadata_dict:
        categories_list = [metadata_dict]
    else:
        categories_number = list(metadata_dict["categories"].keys())
        categories_number.sort()
        categories_list = [metadata_dict["categories"][number] for number in categories_number]
        shared_infos = [key for key in metadata_dict.keys() if key != "categories"]
        for category in categories_list:
            for info in shared_infos:
                if info not in category.keys():
                    category[info] = metadata_dict[info]
    return categories_list
